;Beispiel2 zur Anwendung von AssembleIt2
;AssembleIt2 wird nur dazu verwendet, den ausfhrbaren Code zu erstellen
;Dieser Code kann dann als String gespeichert, und per DllCall() aufgerufen werden!
;DllCall ist die schnellte Variante, assemblierten Code auszufhren!

;
;Fr 32Bit-Modus mus gesetzt werden #AutoIt3Wrapper_UseX64=n
;Fr 64Bit-Modus mus gesetzt werden #AutoIt3Wrapper_UseX64=y


#AutoIt3Wrapper_UseX64=n                                              ;32Bit-Modus

#include-once
#include <assembleit2_64.au3>
#include <SliderConstants.au3>


#cs
    Beispielscript um eine mit ASM-Code erstellte Farbpalette in einer Bitmap darzustellen
    Innerhalb von 2 Schleifen werden die Grn- und Blauwerte von Pixeln gesetzt,
    der Rot-Anteil des Pixels wird durch einen Slider in der GUI festgelegt und an den Code bergeben.
    Weiterhin wird die Startadresse der Bitmap an den Code bergeben.

    Vorgestellt sind mehrere Varianten, um ASM-Code mittels Assembleit2 ablaufen zu lassen.

    1. Beispiel zur Verwendung von Macros und in ASM definierten Variablen (kein DllCallAddress!)
    2. Beispiel als auch per DllCallAddress ausfhrbaren Code (sehr schnell!)
    3. Beispiel mit Verwendung von AutoIt-Variablen im ASM-Code (kein DllCallAddress!)

    Der Overhead von mit AssembleIt2 laufenden Scripten betrgt nur eine Handvoll Millisekunden (2-3ms)!
    In dieser Zeit wird der ASM-Code zur Laufzeit assembliert, in den Speicher geschrieben und dort per DllCallAddress ausgefhrt.

    AssembleIt2 ist auch in der Lage, den Code zu assemblieren und in einer Variable oder Datei abzuspeichern.
    Diesen Code kann man in einer Struct speichern um ihn per DllCallAddress() direkt im Speicher auszufhren.
    Dies ist die zzt. schnellste Variante, um Assemblerprogramme auszufhren. AssembleIt2 und smtliche dafr
    notwendigen Includes sind nicht mehr ntig!

    Bei extrem aufwendigen Berechnungen bleibt daher abzuwgen, ob man die Vorteile der Verwendung von
    Macros und (AutoIt-)Variablen im Code der Erstellung von mit DllCallAddress()lauffhigen Code vorzieht!


    Vom assemblierten Code wird die Anzahl der Prozessortakte zurckgegeben. Bitte beachten, dass die Laufzeit des Codes
    so gering ist, dass moderne Prozessoren im Sleep-Modus bleiben und daher nur mit minimalem Takt laufen!

#ce



;###################################################################################################
;Beispiel 2 zur Verwendung mit DllCallAddress()
;###################################################################################################

#cs _farbverlauf
    use32

    mov edi,[esp+4]                                                   ;pointer pixel
    mov ecx,[esp+8]                                                   ;sliderwert, roter Farbanteil des Pixels

    rdtsc                                                             ;Ticks1 = edx:eax um Takte zu zhlen ^^
    push eax                                                          ;Ticks1 auf Stack speichern

    align 16
    mov edx,256                                                       ;x=256
    _xschleife:                                                       ;for x=256 to 0  grn

    mov ebx,256                                                       ;y=256
    _yschleife:                                                       ;for y=256 to 0  blau

    ;Pixelbytes AARRGGBB zusammensetzen
    ;mov ah,0xFF                        ;0000AA00  alpha, fr die Darstellung in einer Bitmap nicht erforderlich

    mov ah,cl                                                         ;0000RR00  rot
    shl eax,8                                                         ;AARR0000  8 Bit nach links schieben
    mov ah,dl                                                         ;AARRGG00  grn
    mov al,bl                                                         ;AARRGGBB  blau
    stosd                                                             ;ersetzt die nchsten beiden Zeilen, schreibt EAX an [EDI] und erhht EDI um 4
    ;~     mov [edi],eax                       ;pixel an adresse schreiben
    ;~     add edi,4                           ;nchstes pixel

    sub ebx,1                                                         ;blau=blau-1 , sub setzt im Gegensatz zu dec die Flags,
    jnz _yschleife                                                    ;ist ebx=0? JumpNotZero nach label, das Zero-Flag wird hier abgefragt

    sub edx,1                                                         ;grn=grn-1
    jnz _xschleife                                                    ;Grn<>0? ja, dann nchstes grn (Sprung zum Label)

    rdtsc                                                             ;Ticks= edx:eax aktueller Zhlerstand der Ticks seit Rechnerstart
    pop ebx                                                           ;Ticks wurden am Anfang des Scriptes auf den Stack gespeichert
    sub eax,ebx                                                       ;Ticks=Ticks-Ticks1 Anzahl bentigter Takte seit Start zurckgeben

    ret                                                               ;Ende
#ce



$gui = GUICreate("Beispiel", 296, 256)

$slider = GUICtrlCreateSlider(261, 5, 30, 251, BitOR($TBS_VERT, $TBS_TOP, $TBS_LEFT, $TBS_BOTH, $TBS_NOTICKS, $TBS_ENABLESELRANGE))
GUICtrlSetLimit(-1, 255, 0)

;binary aus assemblercode erstellen (fr dllcalladdress bentigt)
$binarycode = _AssembleIt2("retbinary", "_farbverlauf")               ;gibt nur den assemblierten code zurck
ConsoleWrite('@@ Debug(' & @ScriptLineNumber & ') : $binarycode = ' & $binarycode & @CRLF & '>Error code: ' & @error & @CRLF) ;### Debug Console

;nur fr dllcalladdress() bentigt, den binarycode braucht man nur ein mal erstellen
;$binarycode = "0x8B7C24048B4C24080F31509090909090BA00010000BB0001000088CCC1E00888D488D8AB83EB0175F183EA0175E70F315B29D8C3"

Global $tCodeBuffer = _DllStructCreate64("byte[" & StringLen($binarycode) / 2 - 1 & "]") ;reserve Memory for opcodes
DllStructSetData($tCodeBuffer, 1, $binarycode)                        ;"0x8B7C24048B542408B900000000BB00000000B8FF000000C1E00809D0C1E00809D8C1E00809C8890783C7044381FB0001000075DE4181F90001000075D0C3") ;write opcodes into memory

;~ FileDelete("test.bin")  ;man knnte den assemblierten code auch in Dateien speichern
;~ filewrite("test.bin",$binarycode)  ;binary kann als *.bin-File bspw. in IDA eingeleden und debugged werden

GUISetState()

$DC_gui = _WinAPI_GetDC($gui)                                         ;DeviceContext der GUI holen
Global $ptr, $hbmp                                                    ;pointer auf die pixel, handle bitmap
$DC_bitmap = _CreateNewBmp32(256, 256, $ptr, $hbmp)                   ;bitmap erstellen


While GUIGetMsg() <> -3

    $sliderpos = GUICtrlRead($slider)                                 ;sliderposition

    ;entweder Assembleit benutzen, assembliert code whrend der Laufzeit JIT...
    $t = TimerInit()
    Local $ret = _AssembleIt2("uint_ptr", "_farbverlauf", "ptr", $ptr, "int_ptr", $sliderpos) ;, "int_ptr", 33, "int_ptr", 44, "int_ptr", 55, "int_ptr", 66) ;64Bit-Modus
    ConsoleWrite("_AssembleIt2()       Anzahl Takte = " & ($ret) & "     Zeit [ms] = " & Int(1000 * TimerDiff($t)) / 1000 & @CRLF) ;### Debug Console

    ;...oder den schnelleren Dllcalladdress, hier sieht man den Geschwindigkeitsunterschied!
    $t = TimerInit()
    $ret = DllCallAddress("uint:cdecl", DllStructGetPtr($tCodeBuffer), "ptr", $ptr, "int_ptr", $sliderpos)
    ConsoleWrite("DllCallAddress()     Anzahl Takte = " & $ret[0] & "     Zeit [ms] = " & Int(1000 * TimerDiff($t)) / 1000 & @CRLF & @CRLF) ;### Debug Console
;~     ConsoleWrite('@@ : Takte pro Pixel = ' & $ret[0]/256/256 & @CRLF &@crlf) ;### Debug Console

    _WinAPI_BitBlt($DC_gui, 0, 0, 256, 256, $DC_bitmap, 0, 0, 0xCC0020);$srccopy

WEnd


_DeleteBitmap32($DC_bitmap, $ptr, $hbmp)



Func _CreateNewBmp32($iwidth, $iheight, ByRef $ptr, ByRef $hbmp)      ;erstellt leere 32-bit-Bitmap; Rckgabe $HDC und $ptr und handle auf die Bitmapdaten
    $hcdc = _WinAPI_CreateCompatibleDC(0)                             ;Desktop-Kompatiblen DeviceContext erstellen lassen
    $tBMI = DllStructCreate($tagBITMAPINFO)                           ;Struktur der Bitmapinfo erstellen und Daten eintragen
    DllStructSetData($tBMI, 1, DllStructGetSize($tBMI) - 4)           ;Structgre abzglich der Daten fr die Palette
    DllStructSetData($tBMI, 2, $iwidth)
    DllStructSetData($tBMI, 3, -$iheight)                             ;minus =standard = bottomup
    DllStructSetData($tBMI, 4, 1)
    DllStructSetData($tBMI, 5, 32)                                    ;32 Bit = 4 Bytes => AABBGGRR
    $adib = DllCall('gdi32.dll', 'ptr', 'CreateDIBSection', 'hwnd', 0, 'ptr', DllStructGetPtr($tBMI), 'uint', 0, 'ptr*', 0, 'ptr', 0, 'uint', 0)
    $hbmp = $adib[0]                                                  ;hbitmap handle auf die Bitmap, auch per GDI+ zu verwenden
    $ptr = $adib[4]                                                   ;pointer auf den Anfang der Bitmapdaten, vom Assembler verwendet
    _WinAPI_SelectObject($hcdc, $hbmp)                                ;objekt hbitmap in DC
    Return $hcdc                                                      ;DC der Bitmap zurckgeben
EndFunc                                                               ;==>_CreateNewBmp32

Func _DeleteBitmap32($DC, $ptr, $hbmp)
    _WinAPI_DeleteDC($DC)
    _WinAPI_DeleteObject($hbmp)
    $ptr = 0
EndFunc                                                               ;==>_DeleteBitmap32





























